from PyQt5 import QtWidgets, QtCore, QtGui
import redvypr.gui
import sys
import copy
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget
from redvypr.redvypr_address import RedvyprAddress, RedvyprAddressStr
import pydantic
import datetime
import typing
import numpy as np
import uuid


class config_test_sub(pydantic.BaseModel):
    a: str = 'Hello'
    b: float = 1.0

class config_test_allow_extra(pydantic.BaseModel):
    model_config = {'extra':'allow'}
    c: str = 'Hello'
    d: float = 1.0
class config_test(pydantic.BaseModel):
    """
    Calibration model for a heatflow sensor
    """
    structure_version: str = '1.0'
    calibration_type: typing.Literal['A','one','two'] = 'one'
    parameter: str = 'HF'
    sn: str = '' # The serial number of the sensor
    sensor_model: str = ''  # The sensor model
    coeff: float = np.NaN
    coeff_std: typing.Optional[float] = None
    unit:  str = 'W m-2 mV-1'
    raddr: RedvyprAddressStr = pydantic.Field(default=str(RedvyprAddress('d:test')))
    unit_input: str = 'mV'
    date: datetime.datetime = pydantic.Field(default=datetime.datetime(1970,1,1,0,0,0), description='The calibration date')
    calibration_id: str = pydantic.Field(default_factory=lambda: uuid.uuid4().hex, description='ID of the calibration, can be choosen by the user. ID should be unique')
    comment: typing.Optional[str] = None
    uniontest: typing.Union[str,float] = pydantic.Field(default='fdsf')
    uniontest_nonedit: typing.Union[str, float] = pydantic.Field(default='fdsf',editable=False)
    calibrations: typing.List[typing.Union[float, str]] = pydantic.Field(default=[], description = 'List of sensor calibrations')
    #some_dict: typing.Dict[typing.Union[str,float]] = pydantic.Field(default={},
    some_dict: typing.Dict[str,typing.Union[float, bool, list]] = pydantic.Field(default={},description='Configuration of sensors, keys are their serial numbers')
    some_dict2: typing.Dict[str, typing.Union[config_test_sub,float, bool, list]] = pydantic.Field(default={},
                                                                                  description='Configuration of sensors, keys are their serial numbers')
    bmodellist: typing.List[typing.Union[float, config_test_sub,config_test_allow_extra]] = pydantic.Field(default=[],
                                                                         description='List of floats and pydantic models')

    pyd: config_test_allow_extra = pydantic.Field(default=config_test_allow_extra())
    literal: typing.Literal['a','b','c'] = 'a'
    raddr_literal: typing.Union[pydantic.color.Color, typing.Literal['Hallo','Welt'],RedvyprAddressStr] = pydantic.Field(default=str(RedvyprAddress('d:test')))
    color: pydantic.color.Color = pydantic.Field(default=pydantic.color.Color('red'), description='The color of the line')

class config_test_small(pydantic.BaseModel):
    """
    Calibration model for a heatflow sensor
    """

    uniontest_nonedit: typing.Union[str, float] = pydantic.Field(default='fdsf',editable=False)
    a: float = pydantic.Field(default=1.0, editable=False)
    b: dict = {}
    uniontest_2: typing.Union[str,float] = pydantic.Field(default='fdsf')
    uniontest: typing.Union[str, float] = pydantic.Field(default='fdsf')
    uniontest_3: typing.Union[str, float, config_test_sub] = pydantic.Field(default='fdsf')
    some_dict2: typing.Dict[str, typing.Union[config_test_sub, config_test_allow_extra, float, bool, list]] = pydantic.Field(default={},
                                                                                                    description='Configuration of sensors, keys are their serial numbers')
    bmodellist: typing.List[typing.Union[RedvyprAddressStr, float, config_test_sub, config_test_allow_extra]] = pydantic.Field(default=[],
                                                                                                            description='List of floats and pydantic models')
    parameter: typing.Dict[str, typing.Annotated[typing.Union[float, list, RedvyprAddressStr], 'Hallo']] = pydantic.Field(default={})
    raddr: RedvyprAddressStr = pydantic.Field(default=str(RedvyprAddress('d:test')))
    str_format: typing.Dict[str,typing.Union[float,typing.Literal['float', 'int', 'str']]] = pydantic.Field(default={'data_float':'float'})


configtest = config_test_small()

def main():
    app = QtWidgets.QApplication(sys.argv)
    screen = app.primaryScreen()
    #print('Screen: %s' % screen.name())
    size = screen.size()
    #print('Size: %d x %d' % (size.width(), size.height()))
    rect = screen.availableGeometry()
    width = int(rect.width()*4/5)
    height = int(rect.height()*2/3)

    widget = QtWidgets.QWidget()
    layout = QtWidgets.QVBoxLayout(widget)

    #configtree = pydanticConfigWidget(config=configtest, config_location='right')
    # Set the size
    configtree = pydanticConfigWidget(config=configtest, config_location='right', show_editable_only=False)

    layout.addWidget(configtree)
    widget.resize(1000, 800)
    widget.show()
    sys.exit(app.exec_())



if __name__ == '__main__':
    main()



import ast
import copy
import os
import time
import datetime
import logging
import queue
import sys
import yaml
from PyQt5 import QtWidgets, QtCore, QtGui
import qtawesome
import inspect
import uuid
import re
from pyqtconsole.console import PythonConsole
from pyqtconsole.highlighter import format

import redvypr.widgets.redvyprSubscribeWidget
# Import redvypr specific stuff
from redvypr.widgets.standard_device_widgets import displayDeviceWidget_standard, redvypr_deviceInitWidget, RedvyprDeviceWidget_simple, RedvyprDeviceWidget_startonly
#from redvypr.gui import datastreamWidget # Do we need this?
import redvypr.gui as gui
from redvypr.version import version
import redvypr.files as files
import redvypr
import faulthandler
faulthandler.enable()

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('redvypr_main_widget')
logger.setLevel(logging.INFO)

_logo_file = files.logo_file
_icon_file = files.icon_file


class TabGroupButton(QtWidgets.QPushButton):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
#
#
#
# redvyprwidget
#
#
#

class redvyprWidget(QtWidgets.QWidget):
    """This is the main widget of redvypr.

    """

    def __init__(self, width=None, height=None, config=None, hostname='redvypr', loglevel=None, redvypr_device_scan=None):
        """ Args:
            width:
            height:
            config: Either a string containing a path of a yaml file, or a list with strings of yaml files
        """
        super(redvyprWidget, self).__init__()
        self.setGeometry(50, 50, 500, 300)
        # global loglevel
        if loglevel is not None:
            logger.debug('Setting loglevel to global: "{}"'.format(loglevel))
            logger.setLevel(loglevel)
        else:  # config loglevel
            try:
                logger.setLevel(config.loglevel)
            except:
                logger.debug('Could not set loglevel')

        # Lets create the heart of redvypr
        if config is not None:
            config_tmp = config.model_dump(exclude='devices')
            config_tmp_obj = redvypr.RedvyprConfig(**config_tmp)
        else:
            config_tmp_obj = config

        # Configuration comes later after all widgets are initialized
        self.redvypr = redvypr.Redvypr(hostname=hostname,
                                       config=config_tmp_obj,
                                       redvypr_device_scan=redvypr_device_scan,
                                       loglevel=loglevel)
        self.redvypr.device_path_changed.connect(self.__populate_devicepathlistWidget)
        self.redvypr.device_added.connect(self._add_device_gui)
        # Fill the layout
        self.devicetabs = QtWidgets.QTabWidget()
        self.devicetabs.setMovable(True)
        self.devicetabs.setTabsClosable(True)
        self.devicetabs.tabCloseRequested.connect(self.closeTab)


        # Create home tab
        self.createHomeWidget()
        tab_index = self.devicetabs.addTab(self.__homeWidget, 'Home')
        # Add an icon
        try:
            iconname = self.redvypr.config.gui_home_icon
            if iconname == 'redvypr':
                home_icon = QtGui.QIcon(_icon_file)
            else:
                home_icon = qtawesome.icon(iconname)
            logger.debug('Found icon for home')
            self.devicetabs.setTabIcon(tab_index, home_icon)
        except:
            device_icon = None

        # A timer to gather all the data from the devices
        self.devicereadtimer = QtCore.QTimer()
        self.devicereadtimer.timeout.connect(self.readguiqueue)
        self.devicereadtimer.start(100)
        # self.devicereadtimer.start(500)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addWidget(self.devicetabs)
        if ((width is not None) and (height is not None)):
            self.resize(int(width), int(height))

        # TODO, configuration was done earlier, check if this is still ok
        self.redvypr.add_devices_from_config(config)
        # Update hostinformation widgets
        self.__update_hostinfo_widget__()
        self.__populate_devicepathlistWidget()

    def createHomeWidget(self):
        """
        Creates the home widget with the basic information/control/config functionality
        :return:
        """
        self.__homeWidget = QtWidgets.QTabWidget()
        self.__homeWidget_layout = QtWidgets.QVBoxLayout(self.__homeWidget)
        self.__deviceTableWidget = gui.deviceTableWidget(redvyprWidget=self)
        #font = QtGui.QFont('Arial', 20)
        font = QtGui.QFont('Arial')
        font.setBold(True)
        self.__mainLabel = QtWidgets.QLabel('Host information')
        self.__mainLabel.setFont(font)
        self.__mainLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.__homeWidget_layout.addWidget(self.__mainLabel)
        # The configuration of the redvypr
        self.create_devicepathwidget()
        self.create_statuswidget_compact()
        # self.devicetabs.addTab(self.__devicepathwidget,'Status')
        self.__homeWidget_layout.addWidget(self.__statuswidget_compact)

        # Device path
        self.__deviceAddButton = QtWidgets.QPushButton('Add device')
        self.__deviceAddButton.clicked.connect(self.open_add_device_widget)
        self.__deviceAddButton.setFont(font)
        #self.__groupAddButton = QtWidgets.QPushButton('Add group')
        #self.__groupAddButton.clicked.connect(self.__add_group_clicked)
        #self.__groupAddButton.setFont(font)
        self.__homeWidget_layout.addWidget(self.__deviceAddButton)
        #self.__homeWidget_layout.addWidget(self.__groupAddButton)
        self.__homeWidget_layout.addWidget(self.__deviceTableWidget)
        ## Configure button
        #self.__homeWidget_layout.addWidget(self.__host_config_btn)

    def __add_group_clicked(self):
        funcname = __name__ + '.__add_group_clicked():'
        logger.debug(funcname)
        # Create a test group tab
        tab_index = self.devicetabs.addTab(QtWidgets.QWidget(), '')
        tabbar = self.devicetabs.tabBar()
        ind_tab = self.devicetabs.count()
        button = QtWidgets.QPushButton("G G G")
        button.setFixedSize(100, 30)  #
        menu = QtWidgets.QMenu()
        action1 = QtWidgets.QAction("Option 1", self)
        action2 = QtWidgets.QAction("Option 2", self)
        menu.addAction(action1)
        menu.addAction(action2)
        button.setMenu(menu)
        print('Ind tab', ind_tab)
        tabbar.setTabButton(ind_tab - 1, QtWidgets.QTabBar.RightSide, button)
        self.bbbbb = button

    def open_ipwidget(self):
        pass
        #self.ipwidget = redvypr_ip_widget()

    def open_console(self):
        """ Opens a pyqtconsole console widget

        """
        if True:
            width = 800
            height = 500
            # Console
            self.console = PythonConsole(formats={
                'keyword': format('darkBlue', 'bold')
            })
            self.console.setWindowIcon(QtGui.QIcon(_icon_file))
            self.console.setWindowTitle("redvypr console")
            self.console.push_local_ns('redvypr_widget', self)
            self.console.push_local_ns('redvypr', self.redvypr)
            self.console.resize(width, height)
            self.console.show()

            self.console.eval_queued()

        # self.devicetabs.addTab(self.console,'Console')

    def renamedevice(self, oldname, name):
        """ Renames a devices
        """
        funcname = 'renamedevice()'
        for dev in self.redvypr.devices:
            devname = dev['device'].name
            if (devname == name):  # Found a device with that name already, lets do nothging
                logger.debug(funcname + ': Name already in use. Will not rename')
                return False

        for dev in self.redvypr.devices:
            devname = dev['device'].name
            if (devname == oldname):  # Found the device, lets rename it
                dev['device'].change_name(name)
                widget = dev['widget']
                for i in range(self.devicetabs.count()):
                    if (self.devicetabs.widget(i) == widget):
                        self.devicetabs.setTabText(i, name)
                        break

                break

        return True

    def readguiqueue(self):
        """This periodically called function reads the guiqueue and calls
        the widgets of the devices update function (if they exist)

        """
        # Update devices
        for devicedict in self.redvypr.devices:
            device = devicedict['device']
            if True:
                # Feed the data into the modules/functions/objects and
                # let them treat the data
                for i, (guiqueue, gui_widget) in enumerate(devicedict['guiqueues']):
                    while True:
                        try:
                            data = guiqueue.get(block=False)
                        except Exception as e:
                            # print('Exception gui',e)
                            break

                        # Updating the widget, if existing
                        try:
                            gui_widget.update_data(data)
                        except Exception as e:
                            break
                            # logger.exception(e)

    def readguiqueue_legacy(self):
        """This periodically called function reads the guiqueue and calls
        the widgets of the devices update function (if they exist)

        """
        # Update devices
        for devicedict in self.redvypr.devices:
            device = devicedict['device']
            if True:
                # Feed the data into the modules/functions/objects and
                # let them treat the data
                for i, guiqueue in enumerate(devicedict['guiqueue']):
                    while True:
                        try:
                            data = guiqueue.get(block=False)
                        except Exception as e:
                            # print('Exception gui',e)
                            break
                        try:
                            devicedict['gui'][i].update(data)
                        except Exception as e:
                            break
                            # logger.exception(e)

    def load_config(self):
        """ Loads a configuration file
        """
        funcname = self.__class__.__name__ + '.load_config():'
        logger.debug(funcname)
        conffile, _ = QtWidgets.QFileDialog.getOpenFileName(self, "QFileDialog.getOpenFileName()", "",
                                                            "Yaml Files (*.yaml);;All Files (*)")
        if conffile:
            self.redvypr.parse_configuration(conffile)

    def save_config(self):
        """ Saves a configuration file
        """
        funcname = self.__class__.__name__ + '.save_config():'
        logger.debug(funcname)
        config = self.redvypr.get_config()
        data_save = config.model_dump()
        #print('Data save',data_save)
        if True:
            tstr = datetime.datetime.now().strftime('%Y-%m-%d_%H%M%S')
            fname_suggestion = 'config_' + self.redvypr.hostinfo['hostname'] + '_' + tstr + '.yaml'

            fname_full, _ = QtWidgets.QFileDialog.getSaveFileName(self, "QFileDialog.getSaveFileName()",
                                                                  fname_suggestion,
                                                                  "Yaml Files (*.yaml);;All Files (*)")
            #print('fname',fname_full)
            if fname_full:
                logger.debug('Saving to file {:s}'.format(fname_full))
                with open(fname_full, 'w') as fyaml:
                    yaml.dump(data_save, fyaml)

    def open_add_device_widget(self):
        """
        Opens a widget to let the user add redvypr devices
        """
        app = QtWidgets.QApplication.instance()
        screen = app.primaryScreen()
        # print('Screen: %s' % screen.name())
        size = screen.size()
        # print('Size: %d x %d' % (size.width(), size.height()))
        rect = screen.availableGeometry()
        # print('Available: %d x %d' % (rect.width(), rect.height()))
        self.add_device_widget = gui.redvyprAddDeviceWidget(redvypr=self.redvypr)
        self.add_device_widget.resize(int(rect.width() * 0.75), int(rect.height() * 0.75))
        self.add_device_widget.show()

    def _add_device_gui(self, devicelist):
        """ Function is called via the redvypr.add_device signal and is adding
        all the gui functionality to the device

        """
        funcname = __name__ + '._add_device_gui():'
        logger.debug(funcname)
        devicedict = devicelist[0]
        ind_devices = devicelist[1]
        devicemodule = devicelist[2]
        # The widget shown in the tab
        device = devicedict['device']
        devicewidget = QtWidgets.QWidget()
        devicewidget.device = device  # Add the device to the devicewidget
        devicelayout = QtWidgets.QVBoxLayout(devicewidget)

        # Search first for a redvyprdevicewidget, if one is found, take it, otherwise use a init/display widget pair
        try:
            redvyprdevicewidget = devicemodule.RedvyprDeviceWidget
        except:
            logger.debug(funcname + 'Widget does not have a RedvyprDeviceWidget using init/display combination')
            redvyprdevicewidget = None

        if redvyprdevicewidget is not None:
            initargs = inspect.signature(redvyprdevicewidget.__init__)
            initargs_parameters = dict(initargs.parameters)
            initargs2 = None
            if (RedvyprDeviceWidget_startonly in redvyprdevicewidget.__mro__):
                initargs2 = inspect.signature(redvypr_deviceInitWidget.__init__)
            if (RedvyprDeviceWidget_simple in redvyprdevicewidget.__mro__):
                initargs2 = inspect.signature(redvypr_deviceInitWidget.__init__)
            if initargs2 is not None:
                logger.debug(funcname + 'adding arguments {}'.format(dict(initargs2.parameters)))
                initargs_parameters.update(dict(initargs2.parameters))

            initdict = {}
            if ('device' in initargs_parameters.keys()):
                initdict['device'] = device
            if ('redvypr' in initargs_parameters.keys()):
                initdict['redvypr'] = self.redvypr

            device.redvyprdevicewidget = redvyprdevicewidget(**initdict)
            devicelayout.addWidget(device.redvyprdevicewidget)
            self.redvypr.devices[ind_devices]['guiqueues'][0][1] = device.redvyprdevicewidget
            self.redvypr.devices[ind_devices]['displaywidget'] = device.redvyprdevicewidget
            self.redvypr.devices[ind_devices]['initwidget'] = device.redvyprdevicewidget
        else: # Use the init/displaywidget, using tabs
            # The tab for the init and device tab
            devicetab = QtWidgets.QTabWidget()
            devicetab.setMovable(True)
            devicelayout.addWidget(devicetab)
            # Now add all the widgets to the device
            #
            # Create the init widget
            try:
                deviceinitwidget_bare = devicemodule.initDeviceWidget
            except:
                logger.debug(funcname + 'Widget does not have a deviceinitwidget using standard one')
                # logger.exception(e)
                deviceinitwidget_bare = redvypr_deviceInitWidget  # Use a standard widget


            # Call the deviceinitwidget with extra arguments
            initargs = inspect.signature(deviceinitwidget_bare.__init__)
            initargs_parameters = dict(initargs.parameters)
            logger.debug(funcname + 'Initargs for initwidget {}'.format( initargs_parameters))
            # Check if the parent is a deviceinitwidget
            #print('fdsfsd',deviceinitwidget_bare.__mro__)
            if redvypr_deviceInitWidget in deviceinitwidget_bare.__mro__:
                logger.debug(funcname + ' Child of deviceinitwidet')
                initargs2 = inspect.signature(redvypr_deviceInitWidget.__init__)
                logger.debug(funcname + 'adding arguments {}'.format(dict(initargs2.parameters)))
                initargs_parameters.update(dict(initargs2.parameters))
            else:
                initargs_parameters = dict(initargs.parameters)

            initdict = {}
            if ('device' in initargs_parameters.keys()):
                initdict['device'] = device

            if ('redvypr' in initargs_parameters.keys()):
                initdict['redvypr'] = self.redvypr

            if ('tabwidget' in initargs_parameters.keys()):
                initdict['tabwidget'] = devicetab

            # https://stackoverflow.com/questions/334655/passing-a-dictionary-to-a-function-as-keyword-parameters
            try:
                deviceinitwidget = deviceinitwidget_bare(**initdict)
            except Exception as e:
                logger.warning(funcname + 'Could not add deviceinitwidget because of:')
                logger.exception(e)
                deviceinitwidget = QtWidgets.QWidget()  # Use a standard widget

            # Connect the connect signal with connect_device()
            try:
                logger.debug(funcname + 'Connect signal connected')
                deviceinitwidget.connect.connect(self.connect_device)
            except Exception as e:
                logger.debug(funcname + 'Widget does not have connect signal:' + str(e))

            device.deviceinitwidget = deviceinitwidget
            #
            # Check if we have a widget to display the data
            # Create the displaywidget
            #
            try:
                devicedisplaywidget = devicemodule.displayDeviceWidget
            except Exception as e:
                logger.debug(funcname + 'No displaywidget found for {:s}'.format(str(devicemodule)))
                ## Using the standard display widget
                # devicedisplaywidget = displayDeviceWidget_standard
                devicedisplaywidget = None

            # Add init widget
            try:
                tablabelinit = str(device.device_parameter.gui_tablabel_init)
            except:
                tablabelinit = 'Init'
            # print('Device hallo hallo',device.config)
            # device.config['redvypr_device']['gui_tablabel_status']

            devicetab.addTab(deviceinitwidget, tablabelinit)

            # Devices can have their specific display objects, if one is
            # found, initialize it, otherwise just the init Widget
            if (devicedisplaywidget is not None):
                initargs = inspect.signature(devicedisplaywidget.__init__)
                initdict = {}
                if ('device' in initargs.parameters.keys()):
                    initdict['device'] = device

                if ('tabwidget' in initargs.parameters.keys()):
                    initdict['tabwidget'] = devicetab

                if ('deviceinitwidget' in initargs.parameters.keys()):
                    initdict['deviceinitwidget'] = deviceinitwidget

                # https://stackoverflow.com/questions/334655/passing-a-dictionary-to-a-function-as-keyword-parameters
                devicedisplaywidget_called = devicedisplaywidget(**initdict)
                # Add the widget to the device
                device.devicedisplaywidget = devicedisplaywidget_called
                # Test if the widget has a tabname
                try:
                    tablabeldisplay = devicedisplaywidget_called.tabname
                except:
                    try:
                        tablabeldisplay = str(device.device_parameter.gui_tablabel_display)
                    except:
                        tablabeldisplay = 'Display'

                # Check if the widget has included itself, otherwise add the displaytab
                # This is useful to have the displaywidget add several tabs
                # by using the tabwidget argument of the initdict
                if (devicetab.indexOf(devicedisplaywidget_called)) < 0:
                    devicetab.addTab(devicedisplaywidget_called, tablabeldisplay)
                    # Append the widget to the processing queue

                # Update the first entry of the guiqueue list with the displaywidget
                self.redvypr.devices[ind_devices]['guiqueues'][0][1] = devicedisplaywidget_called
                self.redvypr.devices[ind_devices]['displaywidget'] = devicedisplaywidget_called
                self.redvypr.devices[ind_devices]['initwidget'] = deviceinitwidget
            else:
                self.redvypr.devices[ind_devices]['initwidget'] = deviceinitwidget
                self.redvypr.devices[ind_devices]['displaywidget'] = None


        self.redvypr.devices[ind_devices]['widget'] = devicewidget  # This is the displaywidget
        #
        # Add the devicelistentry to the widget, this gives the full information to the device
        #
        # 22.11.2022 TODO, this needs to be replaced by functional arguments instead of properties
        self.redvypr.devices[ind_devices]['initwidget'].redvyprdevicelistentry = self.redvypr.devices[ind_devices]
        self.redvypr.devices[ind_devices]['initwidget'].redvypr = self.redvypr
        if (len(self.redvypr.devices[ind_devices]['guiqueues']) > 0):
            if self.redvypr.devices[ind_devices]['guiqueues'][0][1] is not None:
                self.redvypr.devices[ind_devices]['guiqueues'][0][1].redvyprdevicelistentry = self.redvypr.devices[ind_devices]
                self.redvypr.devices[ind_devices]['guiqueues'][0][1].redvypr = self.redvypr

        # Get an icon for the device
        try:
            iconname = device.device_parameter.gui_icon
            device_icon = qtawesome.icon(iconname)
            logger.debug(funcname + 'Found icon for device')
        except:
            device_icon = None
        # Add the widget to the devicetab, or as a sole window or hide
        widgetname = device.name
        widgetloc = device.device_parameter.gui_dock
        if widgetloc == 'Window':
            devicewidget.setParent(None)
            devicewidget.setWindowTitle(widgetname)
            devicewidget.show()
        elif widgetloc == 'Hide':
            devicewidget.setParent(None)
            devicewidget.setWindowTitle(widgetname)
            devicewidget.hide()
        else:
            if device_icon is not None:
                self.devicetabs.addTab(devicewidget, device_icon, widgetname)
            else:
                self.devicetabs.addTab(devicewidget, widgetname)
            #self.devicetabs.setCurrentWidget(devicewidget)

        # All set, now call finalizing functions
        # Finalize the initialization by calling a helper function (if exist)
        try:
            deviceinitwidget.finalize_init()
        except Exception as e:
            pass
            #logger.debug(funcname + ':finalize_init():' + str(e))

        try:
            devicedisplaywidget_called.finalize_init()
        except Exception as e:
            pass
            #logger.debug(funcname + ':finalize_init():' + str(e))

    def connect_device_gui(self):
        """ Wrapper for the gui
        """
        # Get the current tab
        curtab = self.devicetabs.currentWidget()
        try:
            device = curtab.device
        except:
            device = None
        self.open_connect_widget(device=device)

    def connect_device(self, device):
        """ Handles the connect signal from devices, called when the connection between the device shall be changed
        """
        logger.debug('Connect clicked')
        if (type(device) == dict):
            device = device['device']
        self.open_connect_widget(device=device)

    def open_connect_widget(self, device=None):
        funcname = __name__ + '.open_connect_widget()'
        logger.debug(funcname + ':' + str(device))
        # self.__con_widget = redvyprConnectWidget(devices=self.redvypr.devices, device=device)
        self.__con_widget = redvypr.widgets.redvyprSubscribeWidget.redvyprSubscribeWidget(redvypr=self.redvypr, device=device)
        self.__con_widget.show()

    def __hostname_changed_click(self):
        hostname, ok = QtWidgets.QInputDialog.getText(self, 'redvypr hostname', 'Enter new hostname:')
        if ok:
            self.redvypr.hostinfo['hostname'] = hostname
            self.__hostname_line.setText(hostname)

    def __update_hostinfo_widget__(self):
        """
        Updates the hostinformation
        Returns:

        """
        funcname = __name__ + '.__update_hostinfo_widget__()'
        print(funcname)

    def __update_status_widget__(self):
        """
        Updates the status information
        Returns:

        """


        if self.redvypr.datadistthread.is_alive() == False:
            self.__status_thread.setText(
                'Datadistribution thread is not running! This is bad, consider restarting redvypr.')
            self.__status_thread.setStyleSheet("QLabel { background-color : white; color : red; }")
        else:
            try:
                npackets = self.redvypr.packets_processed
            except:
                npackets = 0
            if (npackets > 0) and (self.redvypr.dt_avg_datadist > 0):
                packets_pstr = npackets / self.redvypr.dt_avg_datadist
            else:
                packets_pstr = 0.0

            trun = time.time() - self.redvypr.t_thread_start
            npackets_total = self.redvypr.packets_counter
            statusstr = 'Running: {:.0f}s, dt: {:0.5f}s (needed {:0.5f}s, {:6.1f} packets/s), Packets processed {:d}'.format(trun, self.redvypr.dt_datadist, self.redvypr.dt_avg_datadist, packets_pstr, npackets_total)
            self.__status_thread.setText(statusstr)

    def create_statuswidget_compact(self):
        """Creates the statuswidget

        """
        self.redvypr.status_update_signal.connect(self.__update_status_widget__)
        self.redvypr.hostconfig_changed_signal.connect(self.__update_hostinfo_widget__)
        self.__statuswidget_compact = QtWidgets.QWidget()
        layout = QtWidgets.QFormLayout(self.__statuswidget_compact)
        self.__status_thread_label = QtWidgets.QLabel('Data distribution thread:')
        # dt
        self.__status_thread = QtWidgets.QLabel('')
        self.__update_status_widget__()
        layout.addRow(self.__status_thread_label,self.__status_thread)#, self.__status_dtneeded)

        # Hostname
        self.__hostname_label = QtWidgets.QLabel('Hostname:')
        self.__hostname_line = QtWidgets.QLabel('')
        self.__hostname_line.setAlignment(QtCore.Qt.AlignRight)
        self.__hostname_line.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.__hostname_line.setText(self.redvypr.hostinfo['hostname'])
        # UUID
        self.__uuid_label = QtWidgets.QLabel('UUID:')
        self.__uuid_line = QtWidgets.QLabel('')
        self.__uuid_line.setAlignment(QtCore.Qt.AlignRight)
        self.__uuid_line.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.__uuid_line.setText(self.redvypr.hostinfo['uuid'])
        # IP
        self.__ip_label = QtWidgets.QLabel('IP:')
        self.__ip_line = QtWidgets.QLabel('')
        self.__ip_line.setAlignment(QtCore.Qt.AlignRight)
        self.__ip_line.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.__ip_line.setText(self.redvypr.hostinfo['addr'])
        # Configuration
        #self.__host_config_btn = QtWidgets.QPushButton('Configure')
        #self.__host_config_btn.clicked.connect(self.__open_configWidget)
        #self.__host_config_widget = QtWidgets.QWidget()
        #self.__host_config_widget_layout = QtWidgets.QFormLayout(self.__host_config_widget)
        # Configuration widgets for detailed configuration, opened when clicked configure button
        # Change the hostname
        self.__hostinfo_opt_btn = QtWidgets.QPushButton('Edit optional information')
        self.__hostinfo_opt_btn.clicked.connect(self.__hostinfo_opt_changed_click)

        self.__statuswidget_pathbtn = QtWidgets.QPushButton('Edit device path')
        self.__statuswidget_pathbtn.clicked.connect(self.show_devicepathwidget)

        layout.addRow(self.__hostname_label, self.__hostname_line)
        layout.addRow(self.__uuid_label, self.__uuid_line)
        layout.addRow(self.__ip_label, self.__ip_line)
        layout.addRow(self.__ip_label, self.__ip_line)
        #layout.addRow(self.__host_config_label,self.__host_config_btn)
        #self.__host_config_widget_layout.addRow(self.__hostinfo_opt_btn)
        #self.__host_config_widget_layout.addRow(self.__statuswidget_pathbtn)


        logo = QtGui.QPixmap(_logo_file)
        logolabel = QtWidgets.QLabel()
        logolabel.setPixmap(logo)
        # layout.addRow(logolabel)

    def __open_configWidget(self):
        self.__host_config_widget.show()

    def create_statuswidget(self):
        """Creates the statuswidget

        """
        self.redvypr.status_update_signal.connect(self.__update_status_widget__)
        self.redvypr.hostconfig_changed_signal.connect(self.__update_hostinfo_widget__)
        self.__statuswidget = QtWidgets.QWidget()
        layout = QtWidgets.QFormLayout(self.__statuswidget)
        # dt
        self.__status_dt = QtWidgets.QLabel('Distribution time: {:0.5f}s'.format(self.redvypr.dt_datadist))
        self.__status_dtneeded = QtWidgets.QLabel(' (needed {:0.5f}s)'.format(self.redvypr.dt_avg_datadist))

        layout.addRow(self.__status_dt, self.__status_dtneeded)

        # Hostname
        self.__hostname_label = QtWidgets.QLabel('Hostname:')
        self.__hostname_line = QtWidgets.QLabel('')
        self.__hostname_line.setAlignment(QtCore.Qt.AlignRight)
        self.__hostname_line.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.__hostname_line.setText(self.redvypr.hostinfo['hostname'])
        # UUID
        self.__uuid_label = QtWidgets.QLabel('UUID:')
        self.__uuid_line = QtWidgets.QLabel('')
        self.__uuid_line.setAlignment(QtCore.Qt.AlignRight)
        self.__uuid_line.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.__uuid_line.setText(self.redvypr.hostinfo['uuid'])
        # IP
        self.__ip_label = QtWidgets.QLabel('IP:')
        self.__ip_line = QtWidgets.QLabel('')
        self.__ip_line.setAlignment(QtCore.Qt.AlignRight)
        self.__ip_line.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.__ip_line.setText(self.redvypr.hostinfo['addr'])

        # Location
        try:
            location = self.redvypr.metadata['location'].data
        except:
            location = ''

        self.__loc_label = QtWidgets.QLabel('Location:')
        self.__loc_text = QtWidgets.QLineEdit(location)
        self.__loc_text.textold = self.__loc_text.text()  # the old text to check again
        self.__loc_text.editingFinished.connect(self.__hostinfo_opt_changed_text)

        # Description
        try:
            description = self.redvypr.metadata['description'].data
        except:
            description = ''

        self.__desc_label = QtWidgets.QLabel('Description:')
        self.__desc_text = QtWidgets.QLineEdit(description)
        self.__desc_text.textold = self.__desc_text.text()  # the old text to check again
        self.__desc_text.editingFinished.connect(self.__hostinfo_opt_changed_text)

        # Lon
        try:
            lon = self.redvypr.metadata['lon'].data
        except:
            lon = -9999.0

        # Lon/Lat
        try:
            lat = self.redvypr.metadata['lat'].data
        except:
            lat = -9999.0

        self.__lon_label = QtWidgets.QLabel('Longitude:')
        self.__lat_label = QtWidgets.QLabel('Latitude:')
        self.__lon_text = QtWidgets.QDoubleSpinBox()
        self.__lon_text.setMinimum(-9999)
        self.__lon_text.setMaximum(360)
        self.__lon_text.setSingleStep(0.00001)
        self.__lon_text.setDecimals(5)
        self.__lon_text.setValue(lon)
        self.__lon_text.oldvalue = self.__lon_text.value()
        self.__lon_text.editingFinished.connect(self.__hostinfo_opt_changed_text)

        self.__lat_text = QtWidgets.QDoubleSpinBox()
        self.__lat_text.setMinimum(-9999)
        self.__lat_text.setMaximum(90)
        self.__lat_text.setSingleStep(0.00001)
        self.__lat_text.setDecimals(5)
        self.__lat_text.setValue(lat)
        self.__lat_text.oldvalue = self.__lat_text.value()
        self.__lat_text.editingFinished.connect(self.__hostinfo_opt_changed_text)

        # Change the hostname
        self.__hostinfo_opt_btn = QtWidgets.QPushButton('Edit optional information')
        self.__hostinfo_opt_btn.clicked.connect(self.__hostinfo_opt_changed_click)

        self.__statuswidget_pathbtn = QtWidgets.QPushButton('Edit device path')
        self.__statuswidget_pathbtn.clicked.connect(self.show_devicepathwidget)

        layout.addRow(self.__hostname_label, self.__hostname_line)
        layout.addRow(self.__uuid_label, self.__uuid_line)
        layout.addRow(self.__ip_label, self.__ip_line)
        layout.addRow(self.__desc_label, self.__desc_text)
        layout.addRow(self.__loc_label, self.__loc_text)
        layout.addRow(self.__lon_label, self.__lon_text)
        layout.addRow(self.__lat_label, self.__lat_text)
        layout.addRow(self.__hostinfo_opt_btn)
        layout.addRow(self.__statuswidget_pathbtn)

        logo = QtGui.QPixmap(_logo_file)
        logolabel = QtWidgets.QLabel()
        logolabel.setPixmap(logo)
        # layout.addRow(logolabel)

    def __hostinfo_opt_changed_text(self):
        """
        Called when the textedit was done, updates the hostinformation,
        Returns:

        """
        funcname = __name__ + '.__hostinfo_opt_changed_text()'
        print(funcname)
        FLAG_CHANGE = False
        # Location text
        if self.__loc_text.textold == self.__loc_text.text():
            print('Not really a change of the text')
        else:
            self.__loc_text.textold = self.__loc_text.text()
            self.redvypr.metadata['location'].data = self.__loc_text.text()
            FLAG_CHANGE = True

        # Location text
        if self.__desc_text.textold == self.__desc_text.text():
            print('Not really a change of the description text')
        else:
            self.__desc_text.textold = self.__desc_text.text()
            self.redvypr.metadata['description'].data = self.__desc_text.text()
            FLAG_CHANGE = True

        # Longitude
        if self.__lon_text.oldvalue == self.__lon_text.value():
            print('Not really a change of the longitude')
        else:
            self.__lon_text.oldvalue = self.__lon_text.value()
            self.redvypr.metadata['lon'].data = self.__lon_text.value()
            FLAG_CHANGE = True

        # Latitude
        if self.__lat_text.oldvalue == self.__lat_text.value():
            print('Not really a change of the latitude')
        else:
            self.__lat_text.oldvalue = self.__lat_text.value()
            self.redvypr.metadata['lat'].data = self.__lat_text.value()
            FLAG_CHANGE = True

        if FLAG_CHANGE:
            print('Things have changed, lets send a signal')
            try:
                self.__hostinfo_opt_edit.reload_config()
            except:
                pass
            self.redvypr.hostconfig_changed_signal.emit()

    def __hostinfo_opt_changed_click(self):
        """
        Opens a widget that allow to change the optional hostinformation
        Returns:

        """
        # Optional hostinformation
        self.__hostinfo_opt_edit = gui.configWidget(self.redvypr.metadata, loadsavebutton=False,
                                                    redvypr_instance=self.redvypr)

        self.__hostinfo_opt_edit.show()

    def show_devicepathwidget(self):
        """A widget to show the pathes to search for additional devices

        """
        self.__devicepathwidget.show()

    def create_devicepathwidget(self):
        """A widget to show the pathes to search for additional devices

        """
        self.__devicepathwidget = QtWidgets.QWidget()
        self.__devicepathlab = QtWidgets.QLabel('Devicepathes')  # Button to add a path
        self.__deviceaddpathbtn = QtWidgets.QPushButton('Add')  # Button to add a path
        self.__deviceaddpathbtn.clicked.connect(self.adddevicepath)
        self.__devicerempathbtn = QtWidgets.QPushButton('Remove')  # Button to remove a path
        self.__devicerempathbtn.clicked.connect(self.remdevicepath)
        layout = QtWidgets.QFormLayout(self.__devicepathwidget)
        self.__devicepathlist = QtWidgets.QListWidget()
        layout.addRow(self.__devicepathlab)
        layout.addRow(self.__devicepathlist)
        layout.addRow(self.__deviceaddpathbtn, self.__devicerempathbtn)
        self.__populate_devicepathlistWidget()

    def __populate_devicepathlistWidget(self):
        self.__devicepathlist.clear()
        for d in self.redvypr.device_paths:
            itm = QtWidgets.QListWidgetItem(d)
            self.__devicepathlist.addItem(itm)

    def __property_widget(self, device=None):
        """

        Returns:

        """
        w = QtWidgets.QWidget()
        if (device == None):
            props = self.properties
        else:
            props = device.properties

    def adddevicepath(self):
        """Adds a path to the devicepathlist
        """
        folder = QtWidgets.QFileDialog.getExistingDirectory(self, 'Devicepath', '')
        if folder:
            self.redvypr.adddevicepath(folder)

    def remdevicepath(self):
        """Removes the selected device pathes
        """
        ind = self.__devicepathlist.currentRow()
        rempath = self.__devicepathlist.item(ind).text()
        # Remove from the main widget and redraw the whole list (is done by the signal emitted by redvypr)
        self.redvypr.remdevicepath(rempath)

    def closeTab(self, currentIndex):
        """ Closing a device tab and removing the device
        """
        funcname = __name__ + '.closeTab()'
        logger.debug('Closing the tab now')
        currentWidget = self.devicetabs.widget(currentIndex)
        # Search for the corresponding device
        for sendict in self.redvypr.devices:
            if (sendict['widget'] == currentWidget):
                device = sendict['device']
                self.redvypr.rem_device(device)
                # Close the widgets (init/display)
                currentWidget.close()
                # TODO: remove
                ## Info
                #sendict['controlwidget'].close()

                self.devicetabs.removeTab(currentIndex)
                break

    def close_application(self):
        funcname = __name__ + '.close_application():'
        print(funcname + ' Closing ...')
        try:
            self.add_device_widget.close()
        except:
            pass

        for sendict in self.redvypr.devices:
            print(funcname + ' Stopping {:s}'.format(sendict['device'].name))
            sendict['device'].thread_stop()

        time.sleep(1)
        for sendict in self.redvypr.devices:
            try:
                sendict['device'].thread.kill()
            except:
                pass

        print('All stopped, sys.exit()')
        # sys.exit()
        os._exit(1)

    def closeEvent(self, event):
        self.close_application()


#
#
# The main widget
#
#
class redvyprMainWidget(QtWidgets.QMainWindow):
    def __init__(self, width=None,height=None,config=None,hostname=None,loglevel=None,redvypr_device_scan=None):
        super(redvyprMainWidget, self).__init__()
        # self.setGeometry(0, 0, width, height)

        # self.setWindowTitle("redvypr")
        self.setWindowTitle(hostname)
        # Add the icon
        # self.setWindowIcon(QtGui.QIcon(_icon_file))

        self.redvypr_widget = redvyprWidget(config=config,
                                            hostname=hostname,
                                            redvypr_device_scan=redvypr_device_scan,
                                            loglevel=loglevel)
        self.setCentralWidget(self.redvypr_widget)
        quitAction = QtWidgets.QAction("&Quit", self)
        quitAction.setShortcut("Ctrl+Q")
        quitAction.setStatusTip('Close the program')
        quitAction.triggered.connect(self.close_application)

        loadcfgAction = QtWidgets.QAction("&Load", self)
        loadcfgAction.setShortcut("Ctrl+O")
        loadcfgAction.setStatusTip('Load a configuration file')
        loadcfgAction.triggered.connect(self.load_config)

        savecfgAction = QtWidgets.QAction("&Save", self)
        savecfgAction.setShortcut("Ctrl+S")
        savecfgAction.setStatusTip('Saves a configuration file')
        savecfgAction.triggered.connect(self.save_config)

        pathAction = QtWidgets.QAction("&Devicepath", self)
        pathAction.setShortcut("Ctrl+L")
        pathAction.setStatusTip('Edit the device path')
        pathAction.triggered.connect(self.redvypr_widget.show_devicepathwidget)

        deviceAction = QtWidgets.QAction("&Add device", self)
        deviceAction.setShortcut("Ctrl+A")
        deviceAction.setStatusTip('Add a device')
        deviceAction.triggered.connect(self.open_add_device_widget)

        devcurAction = QtWidgets.QAction("&Go to home tab", self)
        devcurAction.setShortcut("Ctrl+H")
        devcurAction.setStatusTip('Go to the home tab')
        devcurAction.triggered.connect(self.goto_home_tab)

        # TODO rename to subscribe
        conAction = QtWidgets.QAction("&Connect devices", self)
        conAction.setShortcut("Ctrl+C")
        conAction.setStatusTip('Connect the input/output datastreams of the devices')
        conAction.triggered.connect(self.connect_device_gui)

        self.statusBar()

        mainMenu = self.menuBar()
        fileMenu = mainMenu.addMenu('&File')
        fileMenu.addAction(loadcfgAction)
        fileMenu.addAction(savecfgAction)
        fileMenu.addAction(pathAction)
        fileMenu.addAction(quitAction)

        deviceMenu = mainMenu.addMenu('&Devices')
        deviceMenu.addAction(devcurAction)
        deviceMenu.addAction(deviceAction)
        deviceMenu.addAction(conAction)

        # Help and About menu
        toolMenu = mainMenu.addMenu('&Tools')
        toolAction = QtWidgets.QAction("&Choose Datastreams ", self)
        toolAction.setStatusTip('Opens a window to choose datastreams from the available devices')
        toolAction.triggered.connect(self.show_deviceselect)
        consoleAction = QtWidgets.QAction("&Open console", self)
        consoleAction.triggered.connect(self.open_console)
        consoleAction.setShortcut("Ctrl+N")
        #IPAction = QtWidgets.QAction("&Network interfaces", self)
        #IPAction.triggered.connect(self.open_ipwidget)
        toolMenu.addAction(toolAction)
        #toolMenu.addAction(IPAction)
        toolMenu.addAction(consoleAction)

        # Help and About menu
        helpAction = QtWidgets.QAction("&About", self)
        helpAction.setStatusTip('Information about the software version')
        helpAction.triggered.connect(self.about)

        helpMenu = mainMenu.addMenu('&Help')
        helpMenu.addAction(helpAction)

        self.resize(width, height)
        self.show()

    def open_console(self):
        self.redvypr_widget.open_console()

    def open_ipwidget(self):
        self.redvypr_widget.open_ipwidget()

    def goto_home_tab(self):
        self.redvypr_widget.devicetabs.setCurrentWidget(self.redvypr_widget.__homeWidget)

    def connect_device_gui(self):
        self.redvypr_widget.connect_device_gui()

    def about(self):
        """
        Opens an "about" widget showing basic information.
        """
        self._about_widget = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(self._about_widget)
        label1 = QtWidgets.QTextEdit()
        label1.setReadOnly(True)
        label1.setText(self.redvypr_widget.redvypr.__platform__)
        font = label1.document().defaultFont()
        fontMetrics = QtGui.QFontMetrics(font)
        textSize = fontMetrics.size(0, label1.toPlainText())
        w = textSize.width() + 10
        h = textSize.height() + 20
        label1.setMinimumSize(w, h)
        label1.setMaximumSize(w, h)
        label1.resize(w, h)
        label1.setReadOnly(True)

        layout.addWidget(label1)
        icon = QtGui.QPixmap(_logo_file)
        iconlabel = QtWidgets.QLabel()
        iconlabel.setPixmap(icon)
        layout.addWidget(iconlabel)
        self._about_widget.show()

    def show_deviceselect(self):
        #self.__deviceselect__ = redvypr.widgets.redvyprAddressWidget.datastreamWidget(
        #    redvypr=self.redvypr_widget.redvypr)

        self.__deviceselect__ = redvypr.widgets.redvyprAddressWidget.datastreamQTreeWidget(
            redvypr=self.redvypr_widget.redvypr)
        self.__deviceselect__.show()

    def open_add_device_widget(self):
        self.redvypr_widget.open_add_device_widget()

    def load_config(self):
        self.redvypr_widget.load_config()

    def save_config(self):
        self.redvypr_widget.save_config()

    def close_application(self):
        self.redvypr_widget.close_application()

        sys.exit()

    def closeEvent(self, event):
        self.close_application()






import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import pydantic
import typing
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('rawdatadisp')
logger.setLevel(logging.DEBUG)

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = False
    subscribes: bool = True
    description: str = 'Displays data as text received from connected devices'
    gui_tablabel_display: str = 'Rawdata display'

class DeviceCustomConfig(pydantic.BaseModel):
    bufsize: int = pydantic.Field(default=10000, description='The buffer size of the text display (the MaximumBlockCount of the QPlainTextEdit)')

redvypr_devicemodule = True

def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start()'
    FLAG_RUN = True
    while FLAG_RUN:
        while (datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
            except:
                data = None

            if (data is not None):
                dataqueue.put(data)

                command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
                if (command =='stop'):
                    logger.debug('Stop command for me: {:s}'.format(str(command)))
                    FLAG_RUN = False
                    break

        time.sleep(0.05)




class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QFormLayout(self)
        self.device   = device
        self.text     = QtWidgets.QPlainTextEdit(self)
        self.clearbtn = QtWidgets.QPushButton('Clear')
        self.clearbtn.clicked.connect(self.cleartext)
        self.scrollchk = QtWidgets.QCheckBox('Scroll to end')
        self.text.setReadOnly(True)
        self.text.setMaximumBlockCount(self.device.custom_config.bufsize)
        layout.addRow(self.text)
        layout.addRow(self.scrollchk,self.clearbtn)
        self.text.insertPlainText("Hello, this is the raw data display device!\n")
        
    def cleartext(self):
        self.text.clear()        

    def update_data(self,data):
        #cursor = QtGui.QTextCursor(self.text.document())
        prev_cursor = self.text.textCursor()
        pos = self.text.verticalScrollBar().value()
        self.text.moveCursor(QtGui.QTextCursor.End)
        self.text.insertPlainText(str(data) + '\n')
        #cursor.setPosition(0)
        #self.text.setTextCursor(prev_cursor)
        if(self.scrollchk.isChecked()):
            self.text.verticalScrollBar().setValue(self.text.verticalScrollBar().maximum())
        else:
            self.text.verticalScrollBar().setValue(pos)
        



import json

from PyQt5 import QtWidgets, QtCore, QtGui
import logging
import sys
import qtawesome
import redvypr.files as files
import redvypr.data_packets as data_packets
from redvypr.redvypr_address import RedvyprAddress


_logo_file = files.logo_file
_icon_file = files.icon_file


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('redvyprAddressWidget')
logger.setLevel(logging.DEBUG)


class RedvyprAddressTreeWidget(QtWidgets.QTreeWidget):
    """ A widget that shows all RedvyprAdresses of a device
    Not done yet
    """
    def __init__(self, device=None):
        """
        """
        super(QtWidgets.QWidget, self).__init__()
        self.device = device
        self.root = self.invisibleRootItem()
        self.setColumnCount(1)
        self.populate_tree()
        self.expandAll()
        self.resizeColumnToContents(0)

    def populate_tree(self):
        funcname = __name__ + '.populate_tree():'
        self.clear()
        addresses = self.device.get_deviceaddresses()

        parent = self.root
        for a in addresses:
            item = QtWidgets.QTreeWidgetItem([str(a)])

        parent.addChild(item)


class RedvyprAddressWidget(QtWidgets.QWidget):
    """ A widget that allows to enter an address
    """
    address_finished = QtCore.pyqtSignal(str)  # Signal notifying that the configuration has changed
    def __init__(self, redvypr_address_str=None, redvypr=None):
        """
        """
        super(QtWidgets.QWidget, self).__init__()
        self.layout = QtWidgets.QFormLayout(self)
        self.__configwidget = QtWidgets.QWidget()
        self.redvypr_address = None
        if redvypr_address_str is None:
            manual_address = None
        elif isinstance(redvypr_address_str,RedvyprAddress):
            manual_address = redvypr_address_str.address_str
        elif isinstance(redvypr_address_str, str):
            manual_address = RedvyprAddress(redvypr_address_str).address_str
        else:
            manual_address = None
        self.__datastreamwidget = datastreamWidget(redvypr=redvypr, showapplybutton=False, manual_address=manual_address, datakeys_expanded=True)
        self.__configwidget_input = self.__datastreamwidget.addressline
        self.layout.addRow(self.__datastreamwidget)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyClicked)
        self.__configwidget_apply.__configType = 'configRedvyprAddressStr'
        self.layout.addRow(self.__configwidget_apply)

    def applyClicked(self):
        funcname = __name__ + '.applyClicked():'
        self._test_input()
        logger.debug(funcname + 'Address: {} ({})'.format(self.redvypr_address,type(self.redvypr_address)))
        self.address_finished.emit(str(self.redvypr_address))
    def _test_input(self):
        """
        Tests if the text in the qlineedit is a valid redvypr address
        :return: RedvyprAddress or None
        """
        addr_str = self.__configwidget_input.text()
        #print('Addr str',addr_str)
        try:
            self.redvypr_address = RedvyprAddress(addr_str)
            self.__configwidget_apply.setEnabled(True)
        except:
            logger.debug('Could not parse address string {}'.format(addr_str),exc_info=True)
            self.redvypr_address = None
            self.__configwidget_apply.setEnabled(False)

        #print('Redvypr address',self.redvypr_address)
        return self.redvypr_address


class RedvyprAddressWidgetSimple(QtWidgets.QWidget):
    """ A widget that allows to enter an address
    """
    address_finished = QtCore.pyqtSignal(str)  # Signal notifying that the configuration has changed
    def __init__(self, redvypr_address_str='/d:*'):
        """
        """
        super(QtWidgets.QWidget, self).__init__()
        self.layout = QtWidgets.QFormLayout(self)
        self.__configwidget = QtWidgets.QWidget()
        self.redvypr_address = None
        self.__configwidget_input = QtWidgets.QLineEdit()
        self.__configwidget_input.editingFinished.connect(self._test_input)
        self.__configwidget_input.setText(redvypr_address_str)  # str(data))

        self.layout.addRow(QtWidgets.QLabel('Enter redvypr address'))
        self.layout.addRow(QtWidgets.QLabel('Address string'), self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyClicked)
        self.__configwidget_apply.__configType = 'configRedvyprAddressStr'
        self.__configwidget_cancel = QtWidgets.QPushButton('Cancel')
        self.layout.addRow(self.__configwidget_apply)
        self.layout.addRow(self.__configwidget_cancel)

    def applyClicked(self):
        self._test_input()
        self.address_finished.emit(str(self.redvypr_address))
    def _test_input(self):
        """
        Tests if the text in the qlineedit is a valid redvypr address
        :return: RedvyprAddress or None
        """
        addr_str = self.__configwidget_input.text()
        #print('Addr str',addr_str)
        try:
            self.redvypr_address = RedvyprAddress(addr_str)
            self.__configwidget_apply.setEnabled(True)
        except:
            logger.debug('Could not parse address string {}'.format(addr_str),exc_info=True)
            self.redvypr_address = None
            self.__configwidget_apply.setEnabled(False)

        #print('Redvypr address',self.redvypr_address)
        return self.redvypr_address




class address_filterWidget(QtWidgets.QWidget):
    filterChanged = QtCore.pyqtSignal()  # Signal notifying if the device path was changed
    def __init__(self, redvypr = None):
        """
        """
        self.redvypr = redvypr
        self.filter_address = RedvyprAddress('*')
        self.filter_on = False
        super(QtWidgets.QWidget, self).__init__()
        self.layout = QtWidgets.QGridLayout(self)
        self.btn_nofilter = QtWidgets.QPushButton('Filter off')
        self.btn_nofilter.setCheckable(True)
        self.btn_nofilter.setChecked(False)
        self.btn_nofilter.clicked.connect(self._onfilter_btn_)

        self.btn_showfilter = QtWidgets.QPushButton('Show Filter')
        self.btn_showfilter.setCheckable(True)
        #self.btn_showfilter.setChecked(True)
        self.btn_showfilter.clicked.connect(self._showfilter_btn_)

        self.filter_widget = QtWidgets.QWidget()
        self.filter_layout = QtWidgets.QFormLayout(self.filter_widget)
        self.btn_datakeyfilter = QtWidgets.QPushButton('Datakey')
        self.line_datakeyfilter = QtWidgets.QLineEdit(self.filter_address.datakey)
        self.btn_packetidfilter = QtWidgets.QPushButton('Packet Id')
        self.line_packetidfilter = QtWidgets.QLineEdit(self.filter_address.packetid)
        self.btn_devicefilter = QtWidgets.QPushButton('Device')
        self.line_devicefilter = QtWidgets.QLineEdit(self.filter_address.devicename)
        self.btn_publishingdevicefilter = QtWidgets.QPushButton('Publishing device')
        self.line_publishingdevicefilter = QtWidgets.QLineEdit(self.filter_address.publisher)
        self.btn_hostfilter = QtWidgets.QPushButton('Redvypr host')
        self.line_hostfilter = QtWidgets.QLineEdit(self.filter_address.hostname)

        buttons = [self.btn_datakeyfilter, self.btn_packetidfilter,
                   self.btn_devicefilter, self.btn_publishingdevicefilter,
                   self.btn_hostfilter]
        for b in buttons:
            b.clicked.connect(self.__open_filterChoiceWidget)
            if redvypr is None:
                b.setEnabled(False)

        lineedits = [self.line_datakeyfilter, self.line_packetidfilter,
                     self.line_devicefilter, self.line_publishingdevicefilter,
                     self.line_hostfilter]

        for l in lineedits:
            l.editingFinished.connect(self.__update_address_from_lineedits)

        self.line_filterstr = QtWidgets.QLineEdit(self.filter_address.get_str())

        self.filter_layout.addRow(self.btn_datakeyfilter,self.line_datakeyfilter)
        self.filter_layout.addRow(self.btn_packetidfilter, self.line_packetidfilter)
        self.filter_layout.addRow(self.btn_devicefilter,self.line_devicefilter)
        self.filter_layout.addRow(self.btn_publishingdevicefilter, self.line_publishingdevicefilter)
        self.filter_layout.addRow(self.btn_hostfilter, self.line_hostfilter)
        self.filter_layout.addRow(self.line_filterstr)

        self.filter_widget.hide()
        self.layout.addWidget(self.btn_nofilter,0,0)
        self.layout.addWidget(self.btn_showfilter,0,1)
        self.layout.addWidget(self.filter_widget,1,0,1,2)

    def __open_filterChoiceWidget(self):
        """
        Opens a widget to let the user choose available choices
        """
        self.__filterChoice = QtWidgets.QWidget()
        self.__filterChoiceLayout = QtWidgets.QVBoxLayout(self.__filterChoice)
        self.__filterChoiceList = QtWidgets.QListWidget()
        self.__filterChoiceApply = QtWidgets.QPushButton('Apply')
        self.__filterChoiceApply.clicked.connect(self.__filterChoiceApplyClicked)
        self.__filterChoiceCancel = QtWidgets.QPushButton('Cancel')
        self.__filterChoiceCancel.clicked.connect(self.__filterChoice.close)
        # Fill the list
        if self.sender() == self.btn_datakeyfilter:
            options = self.redvypr.get_datakeys()
            self.__filterChoiceList.lineedit = self.line_datakeyfilter
        elif self.sender() == self.btn_packetidfilter:
            options = self.redvypr.get_packetids()
            self.__filterChoiceList.lineedit = self.line_packetidfilter
        elif self.sender() == self.btn_devicefilter:
            options = self.redvypr.get_devices(local_object=False)
            self.__filterChoiceList.lineedit = self.line_devicefilter
        elif self.sender() == self.btn_publishingdevicefilter:
            options = self.redvypr.get_devices(local_object=True)
            self.__filterChoiceList.lineedit = self.line_publishingdevicefilter
        elif self.sender() == self.btn_hostfilter:
            options = self.redvypr.get_hosts()
            self.__filterChoiceList.lineedit = self.line_hostfilter
        else:
            options = []

        # Append the wildcard
        options.append('*')

        for o in options:
            self.__filterChoiceList.addItem(str(o))

        self.__filterChoiceLayout.addWidget(self.__filterChoiceList)
        self.__filterChoiceLayout.addWidget(self.__filterChoiceApply)
        self.__filterChoiceLayout.addWidget(self.__filterChoiceCancel)
        self.__filterChoiceList.setSelectionMode(QtWidgets.QListWidget.MultiSelection)
        self.__filterChoice.show()

    def __filterChoiceApplyClicked(self):
        options = self.__filterChoiceList.selectedItems()
        if len(options) == 1:
            option = self.__filterChoiceList.currentItem()
            optionstr = str(option.text())
        elif len(options) > 1:
            optionslist = []
            for o in options:
                optionslist.append(o.text())
            if '*' in optionslist:
                optionstr = '*'
            else:
                optionstr ='{'
                for o in optionslist:
                    optionstr += o + '|'

                optionstr = optionstr[:-1] + '}'
        else:
            return

        logger.debug('Apply {}'.format(optionstr))
        self.__filterChoiceList.lineedit.setText(optionstr)
        self.__update_address_from_lineedits()
        self.__filterChoice.close()

    def __update_address_from_lineedits(self):
        host = self.line_hostfilter.text()
        datakey = self.line_datakeyfilter.text()
        packetid = self.line_packetidfilter.text()
        devicename = self.line_devicefilter.text()
        publisher = self.line_publishingdevicefilter.text()
        self.filter_address = RedvyprAddress(datakey=datakey,
                                             packetid=packetid,
                                             hostname=host,
                                             devicename=devicename,
                                             publisher=publisher)
        #print('Update filteraddress',self.filter_address.get_str())
        self.line_filterstr.setText(self.filter_address.get_str())
        self.filterChanged.emit()

    def _onfilter_btn_(self):
        if self.btn_nofilter.isChecked():
            logger.debug('Will filter')
            self.btn_nofilter.setText('Filter on')
            self.filter_on = True
        else:
            logger.debug('Will NOT filter')
            self.btn_nofilter.setText('Filter off')
            self.filter_on = False

        self.filterChanged.emit()


    def _showfilter_btn_(self):
        logger.debug('Show filter button')
        button = self.sender()
        if button.isChecked():
            self.filter_widget.show()
            self.btn_showfilter.setText('Hide Filter')
        else:
            self.filter_widget.hide()
            self.btn_showfilter.setText('Show Filter')


class datastreamWidget(QtWidgets.QWidget):
    """ Widget that lets the user choose available subscribed devices (if device is not None) and datakeys. This
    devicelock: The user cannot change the device anymore
    """
    device_name_changed = QtCore.pyqtSignal(str)  # Signal notifying if the device path was changed
    apply = QtCore.pyqtSignal(dict)  # Signal notifying if the Apply button was clicked
    datakey_name_changed = QtCore.pyqtSignal(str)  # Signal notifying if the datakey has changed

    def __init__(self, redvypr, device=None, devicename_highlight=None, datakey=None, deviceonly=False,
                 devicelock=False, subscribed_only=True, showapplybutton=True,datastreamstring='',closeAfterApply=True,
                 filter_include=[], datakeys_expanded=True, manual_address=None):
        """
        Args:
            redvypr:
            device:
            devicename_highlight: The devicename that is highlighted in the list
            datakey:
            deviceonly:
            devicelock:
            filter_include: List of RedvyprAdresses the will be checked
            subscribed_only: Show the subscribed devices only
            manual_address: String for the manual address
        """

        super(QtWidgets.QWidget, self).__init__()
        logger.setLevel(logging.DEBUG)
        #logger.debug('HALLLOHALLO')
        self.setWindowIcon(QtGui.QIcon(_icon_file))
        self.closeAfterApply = closeAfterApply
        self.redvypr = redvypr
        self.datakeys_expanded = datakeys_expanded
        self.expandlevel = 3
        self.external_filter_include = filter_include
        self.datastreamstring_orig = datastreamstring
        self.datastreamstring  = datastreamstring
        self.layout = QtWidgets.QGridLayout(self)
        self.deviceonly = deviceonly
        if (devicename_highlight == None):
            self.devicename_highlight = 'Na'
        else:
            self.devicename_highlight = devicename_highlight

        self.device = device
        flag_all_devices = (self.device == None) or (subscribed_only == False)  # All devices or only one device?
        try:
            self.devicename = device.name
        except:
            self.devicename = device
        if (device is not None):
            self.devicenamelabel = QtWidgets.QLabel('Device: ' + self.devicename)
            self.layout.addWidget(self.devicenamelabel)
        else:
            self.devicename = ''

        self.deviceavaillabel = QtWidgets.QLabel('Available devices')


        self.devicelist = QtWidgets.QTreeWidget()  # List of available devices
        self.devicelist.setHeaderLabels(['Datastreams'])
        self.devicelist.setColumnCount(1)
        self.devicelist.itemClicked.connect(self.__device_clicked)

        self.addressline_manual = QtWidgets.QLineEdit()
        self.addressline_manual.setReadOnly(False)
        if manual_address is not None:
            self.addressline_manual.setText(manual_address)
        self.addressline_manual.textChanged.connect(self.__addrManualChanged)

        self.addressline = QtWidgets.QLineEdit()
        self.addressline.setReadOnly(True)

        # A combobox to choose between different styles of the address
        self.addrtype_combo = QtWidgets.QComboBox()  # Combo for the different combination types
        redvypr_addresstypes = RedvyprAddress().get_common_address_formats()
        for t in redvypr_addresstypes:
            self.addrtype_combo.addItem(t)

        #self.addrtype_combo.setCurrentIndex(2)
        self.addrtype_combo.currentIndexChanged.connect(self.__addrtype_changed__)
        self.filterWidget = address_filterWidget(redvypr = redvypr)

        # Expansion level
        expandlayout = QtWidgets.QHBoxLayout()
        self.expandlevel_spin = QtWidgets.QSpinBox()
        self.expandlevel_spin.setValue(self.expandlevel)
        self.expandlevel_spin.valueChanged.connect(self.__expandlevelChanged)
        expandlayout.addWidget(QtWidgets.QLabel('Expansion level'))
        expandlayout.addWidget(self.expandlevel_spin)

        # Add widgets to layout
        self.layout_left = QtWidgets.QVBoxLayout()
        self.layout_right = QtWidgets.QVBoxLayout()
        self.layout_right.addLayout(expandlayout)
        self.layout_right.addWidget(self.filterWidget)
        self.layout_left.addWidget(self.deviceavaillabel)
        self.layout_left.addWidget(self.devicelist)

        self.layout_right.addWidget(QtWidgets.QLabel('Manual Address'))
        self.layout_right.addWidget(self.addressline_manual)
        self.layout_right.addWidget(QtWidgets.QLabel('Address format'))
        self.layout_right.addWidget(self.addrtype_combo)
        self.addresslabel = QtWidgets.QLabel('Address')
        self.layout_right.addWidget(self.addresslabel)
        self.layout_right.addWidget(self.addressline)

        self.layout.addLayout(self.layout_left,0,0)
        self.layout.addLayout(self.layout_right,0,1)

        # The datakeys
        if (deviceonly == False):
            pass

        if True:
            self.buttondone = QtWidgets.QPushButton('Apply')
            self.buttondone.clicked.connect(self.done_clicked)
            self.buttondone.setEnabled(False)
        if (showapplybutton):
            self.layout_right.addWidget(self.buttondone)
        else:
            self.buttondone.hide()

        # Add a stretch
        self.layout_right.addStretch()
        devicelist = []
        self.datakeylist_subscribed = {}
        if self.datakeys_expanded:
            self.__update_devicetree_expanded()
            self.filterWidget.filterChanged.connect(self.__update_devicetree_expanded)
        else:
            self.__update_devicetree()
            self.filterWidget.filterChanged.connect(self.__update_devicetree)

    def __expandlevelChanged(self):
        funcname = __name__ + '.____expandlevelChanged():'
        logger.debug(funcname)
        self.expandlevel = self.expandlevel_spin.value()
        self.__update_devicetree_expanded()
    def __addrManualChanged(self,addrstr):
        funcname = __name__ + '.__addrManualChanged():'
        logger.debug(funcname + " manual address: {}".format(addrstr))
        try:
            self.addressline.datakey_address = RedvyprAddress(addrstr)
            self.buttondone.setEnabled(True)
        except:
            logger.info('fdsf',exc_info=True)
            self.buttondone.setEnabled(False)
            return
        self.addressline.device = self.addressline.datakey_address.devicename
        self.addressline.devaddress = self.addressline.datakey_address.address_str
        self.__addrtype_changed__()

    def __device_clicked(self,item):
        """
        Called when an item in the qtree is clicked
        """
        funcname = __name__ + '__device_clicked()'
        logger.debug(funcname)
        #print('Item',item.iskey)
        if(item.iskey): # If this is a datakey item
            addrtype = self.addrtype_combo.currentText()
            addrstring = item.datakey_address.get_str(addrtype)
            #print('Addresstype', addrtype)
            #print('Address',item.datakey_address)
            #print('Addrstring',addrstring)
            #print('Devstring', item.devaddress)
            self.addressline.setText(addrstring)
            self.addressline.datakey_address = item.datakey_address
            self.addressline.device = item.device
            self.addressline.devaddress = item.devaddress
            self.buttondone.setEnabled(True)

    def __update_devicetree(self):
        if True:
            self.devicelist.clear()
            root = self.devicelist.invisibleRootItem()
            # self.devices_listDevices.addItem(str(device))
            data_provider_all = self.redvypr.get_device_objects(publishes=True, subscribes=False)
            font1 = QtGui.QFont('Arial')
            font1.setBold(True)
            font0 = QtGui.QFont('Arial')

            # Fill the qtreewidget
            # print('data provider',data_provider_all)
            if (data_provider_all is not None):
                for dev in data_provider_all:
                    flag_datastreams = False
                    if dev == self.device:
                        continue

                    logger.debug('Address {}'.format(dev.address))
                    # Check for external filter
                    flag_external_filter = True
                    for addr_include in self.external_filter_include:
                        filter_test = dev.address not in addr_include
                        logger.debug('Testing filter {} with {}'.format(dev.address,addr_include))
                        if filter_test:
                            logger.debug('No filter match for external filter {}'.format(dev.address))
                            flag_external_filter = False

                    if flag_external_filter == False:
                        continue
                    # Check for filter from filter widget
                    if self.filterWidget.filter_on:
                        filter_test = dev.address not in self.filterWidget.filter_address
                        logger.debug(
                            'Testing filter for {} in {}: {}'.format(dev.address, self.filterWidget.filter_address,filter_test))
                        if filter_test:
                            logger.debug('No filter match for {} in {}'.format(dev.address, self.filterWidget.filter_address))
                            continue


                    itm = QtWidgets.QTreeWidgetItem([dev.name])
                    col = QtGui.QColor(220,220,220)
                    itm.setBackground(0, col)
                    itm.device = dev
                    itm.redvypr_address = dev.address

                    itm.iskey = False
                    # Check for forwarded devices
                    if True:
                        devs_forwarded = dev.get_device_info()
                        devkeys = list(devs_forwarded.keys())
                        devkeys.sort()
                        for devaddress in devkeys:
                            datakeys = devs_forwarded[devaddress]['datakeys']
                            if len(datakeys) > 0:
                                flag_datastreams = True
                                devaddress_redvypr = RedvyprAddress(devaddress)
                                if self.filterWidget.filter_on:
                                    filter_test = devaddress_redvypr not in self.filterWidget.filter_address
                                    logger.info(
                                        'Testing filter for forwarded {} in {}: {}'.format(devaddress_redvypr,
                                                                                 self.filterWidget.filter_address,
                                                                                 filter_test))

                                    if filter_test:
                                        logger.debug('No filter match for {}'.format(devaddress_redvypr))
                                        continue
                                addrtype = '/d/'
                                devicename = devaddress_redvypr.devicename
                                itmf = QtWidgets.QTreeWidgetItem([devicename])
                                itmf.setBackground(0, col)
                                itmf.device = dev
                                itmf.redvypr_address = devaddress_redvypr
                                itmf.address_forwarded = devaddress
                                itm.addChild(itmf)
                                itmf.iskey = False

                                #print('Datakeys',datakeys,devs_forwarded[devaddress])
                                # Sort the datakey
                                datakeys.sort()
                                for dkey in datakeys:
                                    itmk = QtWidgets.QTreeWidgetItem([dkey])
                                    itmk.iskey = True
                                    itmk.device = dev
                                    itmk.devaddress = devaddress
                                    itmk.datakey_address = RedvyprAddress(devaddress, datakey=dkey)
                                    if self.filterWidget.filter_on:
                                        if itmk.datakey_address not in self.filterWidget.filter_address:
                                            logger.debug('No filter match for {}'.format(itmk.datakey_address))
                                            continue
                                    itmf.addChild(itmk)

                    if flag_datastreams: # If we have datastreams found, add the itm
                        root.addChild(itm)



            self.devicelist.expandAll()
            self.devicelist.resizeColumnToContents(0)

    def __update_devicetree_expanded(self):
        funcname = __name__ + '.__update_devicetree_expanded():'
        logger.debug(funcname)
        colgrey = QtGui.QColor(220, 220, 220)
        def update_recursive(data_new_key, data_new, parent_item, datakey_construct, expandlevel):
            funcname = __name__ + '.__update_recursive():'
            logger.debug(funcname)
            if self.expandlevel == 0:
                datakey_construct_new = data_new_key
            else:
                datakey_construct_new = datakey_construct + '[' + json.dumps(data_new_key) + ']'

            #print('Hallo',data_new_key, data_new,type(data_new))
            #print('Datakey construct new',datakey_construct_new)
            # Check if we are at an item level that is a datakey to be used as a datastream
            if isinstance(data_new, tuple) or (expandlevel >= self.expandlevel):
                #print('Set',data_new,self.expandlevel)
                if expandlevel >= self.expandlevel:
                    #print('Level reached')
                    addrstr_expanded = datakey_construct_new
                else:
                    addrstr = data_new[0]  # Index 0 of set is the address, index 1 the datatype
                    datakeyaddr = RedvyprAddress(addrstr)
                    # Construct a datakey based on the expansion level
                    dkeys_expanded = datakeyaddr.parsed_addrstr_expand['datakeyentries_str']
                    #print('Datakeyaddr', datakeyaddr)
                    #print('expanded datakeys', datakeyaddr.parsed_addrstr_expand['datakeyentries_str'])
                    if datakeyaddr.parsed_addrstr_expand['datakeyeval']:
                        addrstr_expanded = ''
                        for iexpand in range(len(dkeys_expanded)):
                            if iexpand < self.expandlevel:
                                addrstr_expanded += '[' + dkeys_expanded[iexpand] + ']'

                        #print('Addresstr expanded',addrstr_expanded)
                    else:
                        addrstr_expanded = addrstr

                #print('Addresstr expanded',addrstr_expanded)
                itmk = QtWidgets.QTreeWidgetItem([addrstr_expanded])
                itmk.iskey = True
                itmk.device = dev
                itmk.devaddress = devaddress
                #print('Creating address with devaddress',devaddress)
                #print('Creating address with devaddress parsed', devaddress.parsed_addrstr)
                #print('Creating address with datakey', addrstr)
                itmk.datakey_address = RedvyprAddress(devaddress, datakey=addrstr_expanded)
                #print('Address',itmk.datakey_address)
                #print('Address parsed', itmk.datakey_address.parsed_addrstr)
                if self.filterWidget.filter_on:
                    test_filter = itmk.datakey_address not in self.filterWidget.filter_address
                    logger.debug('Testing (@tuple): {} not in {}: {}'.format(itmk.datakey_address,
                                                                             self.filterWidget.filter_address,
                                                                             test_filter))
                    if test_filter:
                        logger.debug('No filter match for {}'.format(itmk.datakey_address))
                    else:
                        parent_item.addChild(itmk)
                else:
                    parent_item.addChild(itmk)

            elif isinstance(data_new, list):
                itmk = QtWidgets.QTreeWidgetItem([str(data_new_key)])
                itmk.setBackground(0, colgrey)
                parent_item.addChild(itmk)
                for data_new_index, data_new_item in enumerate(data_new):
                    update_recursive(data_new_index, data_new_item, parent_item=itmk, datakey_construct=datakey_construct_new, expandlevel=expandlevel+1)

            elif isinstance(data_new, dict):
                itmk = QtWidgets.QTreeWidgetItem([data_new_key])
                itmk.setBackground(0, colgrey)
                parent_item.addChild(itmk)
                for data_new_key in data_new.keys():
                    update_recursive(data_new_key, data_new[data_new_key], parent_item=itmk, datakey_construct=datakey_construct_new, expandlevel=expandlevel+1)

        if True:
            self.devicelist.clear()
            root = self.devicelist.invisibleRootItem()
            # self.devices_listDevices.addItem(str(device))
            data_provider_all = self.redvypr.get_device_objects(publishes=True, subscribes=False)
            font1 = QtGui.QFont('Arial')
            font1.setBold(True)
            font0 = QtGui.QFont('Arial')

            # Fill the qtreewidget
            #print('data provider',data_provider_all)
            if (data_provider_all is not None):
                for dev in data_provider_all:
                    flag_datastreams = False
                    if dev == self.device:
                        continue

                    #print('Device {}'.format(dev.name))
                    #print('Address', dev.address)
                    # Check for external filter
                    flag_external_filter = True
                    for addr_include in self.external_filter_include:
                        test_include = dev.address not in addr_include
                        logger.debug('Testing {} not in {}: {}'.format(dev.address, addr_include, test_include))
                        if test_include:
                            #print('No filter match for external filter', dev.address)
                            flag_external_filter = False

                    if flag_external_filter == False:
                        continue
                    # Check for filter from filter widget
                    if self.filterWidget.filter_on:
                        test_filter = dev.address not in self.filterWidget.filter_address
                        test_filter_sub = True
                        if test_filter == True:
                            # Test all devices of publisher in brute force and check if one of them fits
                            devs_forwarded = dev.get_device_info()
                            for devaddress in devs_forwarded:
                                datakey_dict = devs_forwarded[devaddress]['datakeys_expanded']
                                #print('Datakeys', datakey_dict)
                                devaddress_redvypr = RedvyprAddress(devaddress)
                                if devaddress_redvypr in self.filterWidget.filter_address:
                                    test_filter_sub = False
                                    #print('Filter match for ', devaddress_redvypr)
                                    continue

                        logger.debug('Testing {} not in {}: {}'.format(dev.address, self.filterWidget.filter_address, test_filter))
                        if test_filter and test_filter_sub:
                            #print('No filter match for ', dev.address)
                            continue

                    itm = QtWidgets.QTreeWidgetItem([dev.name])
                    col = QtGui.QColor(220, 220, 220)
                    itm.setBackground(0, col)
                    itm.device = dev
                    itm.redvypr_address = dev.address

                    itm.iskey = False
                    # Check for forwarded devices
                    if True:
                        devs_forwarded = dev.get_device_info()
                        devkeys = list(devs_forwarded.keys())
                        devkeys.sort()
                        for devaddress in devkeys:
                            datakey_dict = devs_forwarded[devaddress]['datakeys_expanded']
                            #print('Datakeys',datakey_dict)

                            #fdsfdsf
                            flag_datastreams = True
                            devaddress_redvypr = RedvyprAddress(devaddress)
                            if self.filterWidget.filter_on:
                                if devaddress_redvypr not in self.filterWidget.filter_address:
                                    #print('No filter match for ', devaddress_redvypr)
                                    continue
                            addrtype = '/d/i/'
                            #print('Hallo', devaddress_redvypr, devaddress_redvypr.get_str())
                            devicestr = devaddress_redvypr.devicename
                            # TODO, this should be defined in the configuration of the widget
                            #devicestr = devaddress_redvypr.get_str(addrtype)
                            devicestr = devaddress_redvypr.get_str()
                            itmf = QtWidgets.QTreeWidgetItem([devicestr])
                            itmf.setBackground(0, col)
                            itmf.device = dev
                            itmf.redvypr_address = devaddress_redvypr
                            itmf.address_forwarded = devaddress
                            itm.addChild(itmf)
                            itmf.iskey = False

                            for key in datakey_dict.keys():
                                data_new = datakey_dict[key]
                                datakey_construct_new = ''
                                update_recursive(key, data_new, parent_item=itmf, datakey_construct=datakey_construct_new, expandlevel=0)

                    if flag_datastreams:  # If we have datastreams found, add the itm
                        root.addChild(itm)

            self.devicelist.expandAll()
            self.devicelist.resizeColumnToContents(0)

    def __addrtype_changed__(self):
        """ Update the datakeylist whenever the device was changed
        """
        funcname = __name__ + '.__addrtype_changed__():'
        logger.debug(funcname)
        addrtype = self.addrtype_combo.currentText()
        try:
            addrstring =  self.addressline.datakey_address.get_str(addrtype)
            self.buttondone.setEnabled(True)
        except:
            addrstring = ''
            self.buttondone.setEnabled(False)

        self.addressline.setText(addrstring)

    def done_clicked(self):
        funcname = __name__ + '.done_clicked():'
        addrformat = self.addrtype_combo.currentText()
        device = self.addressline.device
        device_address = self.addressline.devaddress
        datastream_address = self.addressline.datakey_address
        datastream_str = datastream_address.get_str(addrformat)

        signal_dict = {'device': device, 'device_address':device_address,'datastream_str': datastream_str,'datastream_address':datastream_address,'address_format':addrformat}
        signal_dict['addrformat'] = self.addressline.datakey_address
        print(funcname + 'Signal dict {}'.format(signal_dict))
        self.apply.emit(signal_dict)
        if self.closeAfterApply:
            self.close()



class datastreamsWidget(datastreamWidget):
    """ Widget that lets the user choose several datastreams
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args,**kwargs)
        self.devicelist.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.datastreamtable = QtWidgets.QTableWidget()
        self.layout.addWidget(self.datastreamtable,0,2)
        self.layout.removeWidget(self.buttondone)
        self.buttondone.clicked.disconnect(self.done_clicked)
        self.buttondone.clicked.connect(self.apply_clicked_datastreams)

        self.layout_right.removeWidget(self.addresslabel)
        self.addresslabel.hide()
        self.layout_right.removeWidget(self.addressline)
        self.addressline.hide()
        self.addrtype_combo.currentIndexChanged.disconnect(self.__addrtype_changed__)
        self.addrtype_combo.currentIndexChanged.connect(self.update_datastreamtable)

        iconname='ei.remove'
        icon = qtawesome.icon(iconname)
        self.button_rem = QtWidgets.QPushButton('Remove')
        self.button_rem.setIcon(icon)
        self.button_rem.clicked.connect(self.rem_datastreams)
        iconname='ei.caret-right'
        icon = qtawesome.icon(iconname)
        self.button_add = QtWidgets.QPushButton('Add')
        self.button_add.setIcon(icon)
        self.button_add.clicked.connect(self.add_datastreams)
        self.layout_right.addWidget(self.button_rem)
        self.layout_right.addWidget(self.button_add)
        self.layout.addWidget(self.buttondone,1,0,1,3)
        self.addresses_choosen = []
        self.update_datastreamtable()

    def apply_clicked_datastreams(self):
        funcname = __name__ + '.apply_clicked_datastreams()'
        logger.debug(funcname)
        addresses_choosen = []
        for irow, raddr in enumerate(self.addresses_choosen):
            addrtype = self.addrtype_combo.currentText()
            addrstr = raddr.get_str(addrtype)  # Here a format would be nice
            addresses_choosen.append(RedvyprAddress(addrstr))

        signal_dict = {'addresses':addresses_choosen}
        print('Signal dict',signal_dict)
        self.apply.emit(signal_dict)
        if self.closeAfterApply:
            self.close()

    def update_datastreamtable(self):
        self.datastreamtable.clear()
        nrows = len(self.addresses_choosen)
        self.datastreamtable.setRowCount(nrows)
        self.datastreamtable.setColumnCount(1)
        for irow,raddr in enumerate(self.addresses_choosen):
            addrtype = self.addrtype_combo.currentText()
            addrstr = raddr.get_str(addrtype) # Here a format would be nice
            item = QtWidgets.QTableWidgetItem(addrstr)
            item.datakey_address = raddr
            self.datastreamtable.setItem(irow,0, item)

        self.datastreamtable.setHorizontalHeaderLabels(['Address'])
        self.datastreamtable.resizeColumnsToContents()
        if len(self.addresses_choosen)>0:
            self.buttondone.setEnabled(True)
        else:
            self.buttondone.setEnabled(False)

    def rem_datastreams(self):
        funcname = __name__ + '.rem_datastreams():'
        logger.debug(funcname)
        for item in self.datastreamtable.selectedItems():
            print("selectedItem", item.text())
            self.addresses_choosen.remove(item.datakey_address)

        self.update_datastreamtable()

    def add_datastreams(self):
        funcname = __name__ + '.add_datastreams():'
        logger.debug(funcname)
        items = self.devicelist.selectedItems()
        for i,item in enumerate(items):
            #print(i,item.text(0))
            if item.iskey:
                print('Item {} is a valid address'.format(item.text(0)))
                if item.datakey_address not in self.addresses_choosen:
                    self.addresses_choosen.append(item.datakey_address)
                else:
                    print('Address is existing already')

            else:
                print('Item {} is not a datastream'.format(item.text(0)))


        #print('Addresses',self.addresses_choosen)
        self.update_datastreamtable()


class datastreamQTreeWidget(QtWidgets.QWidget):
    """ Widget shows all datastreams in a QTree style
    """
    def __init__(self, redvypr, device=None, filter_include=[], headerlabel=''):
        super(QtWidgets.QWidget, self).__init__()
        logger.setLevel(logging.DEBUG)
        self.setWindowIcon(QtGui.QIcon(_icon_file))
        self.redvypr = redvypr
        self.datakeys_expanded = True
        self.expandlevel = 3
        self.external_filter_include = filter_include
        self.layout = QtWidgets.QGridLayout(self)
        self.device = device
        try:
            self.devicename = device.name
        except:
            self.devicename = device
        if (device is not None):
            pass
            #self.devicenamelabel = QtWidgets.QLabel('Device: ' + self.devicename)
            #self.layout.addWidget(self.devicenamelabel)
        else:
            self.devicename = ''

        self.deviceavaillabel = QtWidgets.QLabel('Available devices')
        self.devicelist = QtWidgets.QTreeWidget()  # List of available devices
        self.devicelist.setColumnCount(1)
        self.devicelist.setHeaderLabels([headerlabel])
        self.devicelist.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.devicelist.header().setStretchLastSection(False)
        #self.devicelist.setAutoScroll(True)


        self.filterWidget = address_filterWidget(redvypr = redvypr)
        # Expansion level
        expandlayout = QtWidgets.QHBoxLayout()
        self.expandlevel_spin = QtWidgets.QSpinBox()
        self.expandlevel_spin.setValue(self.expandlevel)
        self.expandlevel_spin.valueChanged.connect(self.__expandlevelChanged)
        expandlayout.addWidget(QtWidgets.QLabel('Expansion level'))
        expandlayout.addWidget(self.expandlevel_spin)

        # Add widgets to layout
        self.deviceWidget = QtWidgets.QWidget()
        self.layout_left = QtWidgets.QVBoxLayout(self.deviceWidget)
        self.settingsWidget = QtWidgets.QWidget()
        self.layout_right = QtWidgets.QVBoxLayout(self.settingsWidget)
        self.layout_right.addLayout(expandlayout)
        self.layout_right.addWidget(self.filterWidget)
        self.layout_left.addWidget(self.deviceavaillabel)
        self.layout_left.addWidget(self.devicelist)

        # Create a splitter
        # if self.config_location == 'bottom':
        #    sdir = QtCore.Qt.Vertical
        # else:
        #    sdir = QtCore.Qt.Horizontal

        sdir = QtCore.Qt.Vertical
        self.splitter = QtWidgets.QSplitter(sdir)
        self.splitter.addWidget(self.deviceWidget)
        self.splitter.addWidget(self.settingsWidget)

        self.layout.addWidget(self.splitter,0,0)

        ## Add a stretch
        #self.layout_right.addStretch()
        if self.datakeys_expanded:
            self.__update_devicetree_expanded()
            self.filterWidget.filterChanged.connect(self.__update_devicetree_expanded)
        else:
            self.__update_devicetree()
            self.filterWidget.filterChanged.connect(self.__update_devicetree)

    def __expandlevelChanged(self):
        funcname = __name__ + '.____expandlevelChanged():'
        logger.debug(funcname)
        self.expandlevel = self.expandlevel_spin.value()
        self.__update_devicetree_expanded()


    def apply_address_filter(self,device, filter_address):#
        funcname = __name__ + '.apply_address_filter():'
        logger.debug(funcname + 'Testing {} in {} (with subsearch)'.format(device.address, filter_address))
        test_filter = device.address in filter_address
        test_filter_sub = False
        if test_filter == False: # Check if subdevices have a match
            # Test all devices of publisher in brute force and check if one of them fits
            devs_forwarded = device.get_device_info()
            for devaddress in devs_forwarded:
                datakey_dict = devs_forwarded[devaddress]['datakeys_expanded']
                # print('Datakeys', datakey_dict)
                devaddress_redvypr = RedvyprAddress(devaddress)
                if devaddress_redvypr in filter_address:
                    test_filter_sub = True
                    print('Filter match for subsearch:', devaddress_redvypr)
                    break


        if test_filter or test_filter_sub:
            print('Filter match for ', device.address)
            return True
        else:
            print('No filter match for ', device.address)
            return False


    def __update_devicetree_expanded(self):
        funcname = __name__ + '.__update_devicetree_expanded():'
        logger.debug(funcname)
        colgrey = QtGui.QColor(220, 220, 220)
        def update_recursive(data_new_key, data_new, parent_item, datakey_construct, expandlevel):
            funcname = __name__ + '.__update_recursive():'
            #logger.debug(funcname)
            if self.expandlevel == 0:
                datakey_construct_new = data_new_key
            else:
                datakey_construct_new = datakey_construct + '[' + json.dumps(data_new_key) + ']'

            raddress_constructed = RedvyprAddress(devaddress, datakey=datakey_construct_new)
            #print('Hallo',data_new_key, data_new,type(data_new))
            #print('Datakey construct new',datakey_construct_new)
            # Check if we are at an item level that is a datakey to be used as a datastream
            if isinstance(data_new, tuple) or (expandlevel >= self.expandlevel):
                #print('Set',data_new,self.expandlevel)
                if expandlevel >= self.expandlevel:
                    #print('Level reached')
                    addrstr_expanded = datakey_construct_new
                else:
                    addrstr = data_new[0]  # Index 0 of set is the address, index 1 the datatype
                    datakeyaddr = RedvyprAddress(addrstr)
                    # Construct a datakey based on the expansion level
                    dkeys_expanded = datakeyaddr.parsed_addrstr_expand['datakeyentries_str']
                    #print('Datakeyaddr', datakeyaddr)
                    #print('expanded datakeys', datakeyaddr.parsed_addrstr_expand['datakeyentries_str'])
                    if datakeyaddr.parsed_addrstr_expand['datakeyeval']:
                        addrstr_expanded = ''
                        for iexpand in range(len(dkeys_expanded)):
                            if iexpand < self.expandlevel:
                                addrstr_expanded += '[' + dkeys_expanded[iexpand] + ']'

                        #print('Addresstr expanded',addrstr_expanded)
                    else:
                        addrstr_expanded = addrstr

                #print('Addresstr expanded',addrstr_expanded)
                itmk = QtWidgets.QTreeWidgetItem([addrstr_expanded])
                itmk.iskey = True
                itmk.device = dev
                itmk.devaddress = devaddress
                #print('Creating address with devaddress',devaddress)
                #print('Creating address with devaddress parsed', devaddress.parsed_addrstr)
                #print('Creating address with datakey', addrstr)
                itmk.datakey_address = RedvyprAddress(devaddress, datakey=addrstr_expanded)
                itmk.raddress = itmk.datakey_address
                #print('Address',itmk.datakey_address)
                #print('Address parsed', itmk.datakey_address.parsed_addrstr)
                if self.filterWidget.filter_on:
                    test_filter = itmk.datakey_address not in self.filterWidget.filter_address
                    # TODO: Here also the external filter should be checked
                    logger.debug('Testing (@tuple): {} not in {}: {}'.format(itmk.datakey_address,
                                                                             self.filterWidget.filter_address,
                                                                             test_filter))
                    if test_filter:
                        logger.debug('No filter match for {}'.format(itmk.datakey_address))
                    else:
                        parent_item.addChild(itmk)
                else:
                    parent_item.addChild(itmk)

            elif isinstance(data_new, list):
                itmk = QtWidgets.QTreeWidgetItem([str(data_new_key)])
                itmk.setBackground(0, colgrey)
                itmk.raddress = raddress_constructed
                parent_item.addChild(itmk)
                for data_new_index, data_new_item in enumerate(data_new):
                    update_recursive(data_new_index, data_new_item, parent_item=itmk, datakey_construct=datakey_construct_new, expandlevel=expandlevel+1)

            elif isinstance(data_new, dict):
                itmk = QtWidgets.QTreeWidgetItem([data_new_key])
                itmk.setBackground(0, colgrey)
                itmk.raddress = raddress_constructed
                parent_item.addChild(itmk)
                for data_new_key in data_new.keys():
                    update_recursive(data_new_key, data_new[data_new_key], parent_item=itmk, datakey_construct=datakey_construct_new, expandlevel=expandlevel+1)

        if True:
            self.devicelist.clear()
            root = self.devicelist.invisibleRootItem()
            # self.devices_listDevices.addItem(str(device))
            data_provider_all = self.redvypr.get_device_objects(publishes=True, subscribes=False)
            font1 = QtGui.QFont('Arial')
            font1.setBold(True)
            font0 = QtGui.QFont('Arial')

            # Fill the qtreewidget
            #print('data provider',data_provider_all)
            if (data_provider_all is not None):
                for dev in data_provider_all:
                    flag_datastreams = False
                    #if dev == self.device:
                    #    continue

                    #print('Device {}'.format(dev.name))
                    #print('Address', dev.address)
                    # Check for external filter
                    flag_external_filter = True
                    for addr_include in self.external_filter_include:
                        flag_external_filter = self.apply_address_filter(dev,addr_include)
                    if flag_external_filter == False:
                        continue
                    # Check for filter from filter widget
                    if self.filterWidget.filter_on:
                        test_filter = dev.address not in self.filterWidget.filter_address
                        test_filter_sub = True
                        if test_filter == True:
                            # Test all devices of publisher in brute force and check if one of them fits
                            devs_forwarded = dev.get_device_info()
                            for devaddress in devs_forwarded:
                                datakey_dict = devs_forwarded[devaddress]['datakeys_expanded']
                                #print('Datakeys', datakey_dict)
                                devaddress_redvypr = RedvyprAddress(devaddress)
                                #print('a',devaddress_redvypr,self.filterWidget.filter_address)
                                if devaddress_redvypr in self.filterWidget.filter_address:
                                    test_filter_sub = False
                                    #print('Filter match for ', devaddress_redvypr)
                                    continue

                        logger.debug('Testing {} not in {}: {}'.format(dev.address, self.filterWidget.filter_address, test_filter))
                        if test_filter and test_filter_sub:
                            #print('No filter match for ', dev.address)
                            continue

                    itm = QtWidgets.QTreeWidgetItem([dev.name])
                    col = QtGui.QColor(220, 220, 220)
                    itm.setBackground(0, col)
                    itm.device = dev
                    itm.redvypr_address = dev.address
                    itm.raddress = dev.address

                    itm.iskey = False
                    # Check for forwarded devices
                    if True:
                        devs_forwarded = dev.get_device_info()
                        devkeys = list(devs_forwarded.keys())
                        devkeys.sort()
                        for devaddress in devkeys:
                            datakey_dict = devs_forwarded[devaddress]['datakeys_expanded']
                            #print('Datakeys',datakey_dict)

                            #fdsfdsf
                            flag_datastreams = True
                            devaddress_redvypr = RedvyprAddress(devaddress)
                            # Filter the addresses
                            flag_external_filter_match = False
                            for addr_include in self.external_filter_include:
                                if devaddress_redvypr in addr_include:
                                    flag_external_filter_match = True
                                    break

                            # No external filter match, ignore device
                            if not(flag_external_filter_match):
                                continue
                            if self.filterWidget.filter_on:
                                if devaddress_redvypr not in self.filterWidget.filter_address:
                                    #print('No filter match for ', devaddress_redvypr)
                                    continue
                            addrtype = '/d/i/'
                            #print('Hallo', devaddress_redvypr, devaddress_redvypr.get_str())
                            devicestr = devaddress_redvypr.devicename
                            # TODO, this should be defined in the configuration of the widget
                            #devicestr = devaddress_redvypr.get_str(addrtype)
                            devicestr = devaddress_redvypr.get_str()
                            itmf = QtWidgets.QTreeWidgetItem([devicestr])
                            itmf.setBackground(0, col)
                            itmf.device = dev
                            itmf.redvypr_address = devaddress_redvypr
                            itmf.address_forwarded = devaddress
                            itmf.raddress = devaddress_redvypr
                            itm.addChild(itmf)
                            itmf.iskey = False

                            for key in datakey_dict.keys():
                                data_new = datakey_dict[key]
                                datakey_construct_new = ''
                                update_recursive(key, data_new, parent_item=itmf, datakey_construct=datakey_construct_new, expandlevel=0)

                    if flag_datastreams:  # If we have datastreams found, add the itm
                        root.addChild(itm)

            self.devicelist.expandAll()
            #self.devicelist.header().setStretchLastSection(True)
            self.devicelist.resizeColumnToContents(0)






